"""
A series of functions directly obtained from the Trajectory class. These functions allow the user to manually specify
how they would want their data to be processed with more flexibility.

Written by X.M. Wang.

Wang, X.M., & Welsh, T.N. (2023). TAT-HUM: Trajectory Analysis Toolkit for Human Movements in Python.
"""

from typing import Union, Optional
import vg
import numpy as np
from scipy import interpolate
from scipy.signal import butter, filtfilt
from scipy.spatial.transform import Rotation
from skspatial.objects import Plane, Points, Vector
from enum import Enum
from functools import reduce


def check_input_coord_2d(x, y):
    """
    a helper function to check the dimensions of the input x and y coordinates
    """
    if not x.shape == y.shape:
        raise ValueError('input x and y have to be the same shape!')

    if len(x.shape) > 1:
        x = x.squeeze()
        y = y.squeeze()

    return x, y


def check_input_coord(x, y, z):
    """
    a helper function to check the dimensions of the input x, y, and z coordinates
    """
    if not x.shape == y.shape == z.shape:
        raise ValueError('input x, y, and z have to be the same shape!')

    if len(x.shape) > 1:
        x = x.squeeze()
        y = y.squeeze()
        z = z.squeeze()

    return x, y, z


def consecutive(data, step_size=1):
    """ splits the missing data indices into chunks"""
    return np.split(data, np.where(np.diff(data) != step_size)[0] + 1)


def cent_diff(time: np.ndarray, signal: np.ndarray):
    """
    Central difference method to find derivatives.

    :param time: the time vector
    :param signal: the signal to be smoothed
    """
    n_frames = len(time)

    if len(signal) != n_frames:
        raise ValueError('The input signal has to be of the same length as its corresponding time vector!')

    der = np.zeros(n_frames, dtype=float)

    der[0] = (signal[1] - signal[0]) / (time[1] - time[0])
    der[-1] = (signal[-1] - signal[-2]) / (time[-1] - time[-2])

    for i_frame in np.arange(1, n_frames - 1):
        der[i_frame] = (signal[i_frame + 1] - signal[i_frame - 1]) / (
                time[i_frame + 1] - time[i_frame - 1])

    return der


# def find_missing_data(coord: np.ndarray,
#                       missing_data_value=0.):
#     """
#     identify the missing data points using any coordinate
#
#     :param x: position along the x-axis
#     :param y: position along the y-axis
#     :param z: position along the z-axis
#     :param missing_data_value: the default value for missing data points
#     :return: x, y, z, missing_info = {contain_missing, n_missing, missing_ind}
#     """
#
#     missing_ind = np.where(coord == missing_data_value)[0]
#     not_missing_ind = np.where(coord != missing_data_value)[0]
#     contain_missing = len(missing_ind) > 0
#     return contain_missing, missing_ind, not_missing_ind


def fill_missing_data(x: np.ndarray,
                      y: np.ndarray,
                      time: np.ndarray,
                      z: Optional[np.ndarray] = None,
                      missing_data_value=0.):
    """
    identify and fill missing data using Scipy's interp1d function

    :param x: position along the x-axis
    :param y: position along the y-axis
    :param z: position along the z-axis
    :param time: the time vector
    :param missing_data_value: the default value for missing data points
    :return: x, y, z, time, missing_info = {contain_missing, n_missing, missing_ind}
    """

    if z is None:
        x, y = check_input_coord_2d(x, y)
    else:
        x, y, z = check_input_coord(x, y, z)

    missing_ind = np.where(x == missing_data_value)[0]
    not_missing_ind = np.where(x != missing_data_value)[0]

    if len(missing_ind) > 0:

        f_interp = interpolate.interp1d(time[not_missing_ind], x[not_missing_ind], bounds_error=False,
                                        # fill values are for missing values outside the range of x
                                        fill_value=(np.NaN, np.NaN))
        x = f_interp(time)

        f_interp = interpolate.interp1d(time[not_missing_ind], y[not_missing_ind], bounds_error=False,
                                        fill_value=(np.NaN, np.NaN))
        y = f_interp(time)

        if z is not None:
            f_interp = interpolate.interp1d(time[not_missing_ind], z[not_missing_ind], bounds_error=False,
                                            fill_value=(np.NaN, np.NaN))
            z = f_interp(time)

        ind_delete = np.where(np.isnan(x))[0]
        x = np.delete(x, ind_delete)
        y = np.delete(y, ind_delete)
        if z is not None:
            z = np.delete(z, ind_delete)
        time = np.delete(time, ind_delete)

        # identify the consecutive missing data points
        missing_ind_segments = consecutive(missing_ind)
        n_missing_segments = [len(seg) for seg in missing_ind_segments]

        missing_info = {
            'contain_missing': True,
            'n_missing': len(missing_ind),
            'missing_ind': missing_ind,
            'missing_ind_segments': missing_ind_segments,
            'n_missing_segments': n_missing_segments
        }

        if z is None:
            return x, y, time, missing_info
        else:
            return x, y, z, time, missing_info
    else:
        missing_info = {
            'contain_missing': False,
            'n_missing': 0,
            'missing_ind': [],
            'missing_ind_segments': [],
            'n_missing_segments': []
        }

        if z is None:
            return x, y, time, missing_info
        else:
            return x, y, z, time, missing_info


def low_butter(signal, fs, fc, order=2):
    """
    Direct usage of the low-pass Butterworth Filter using library from SciPy.

    :param signal: 1D data to be filtered
    :param fs: sampling frequency
    :param fc: cutoff frequency
    :param order: butterworth order
    :return: filtered signal
    """

    Wn = fc / (fs / 2)
    poly = butter(order, Wn, btype='lowpass', output='ba')  # returns numerator [0] and denominator [1] polynomials
    return filtfilt(poly[0], poly[1], signal.copy())


def find_optimal_cutoff_frequency(signal: np.ndarray,
                                  fs: int,
                                  fc_test: np.ndarray = np.arange(2, 14),
                                  ):
    """
    Identify the optimal cutoff frequency based on the normalized residual autocorrelation (see Cappello, La
    Palombara, & Leardini, 1996)

    :param signal: the signal to be filtered
    :param fs: sampling frequency
    :param fc_test: the range of cutoff frequencies to be tested
    Default range is obtained from Schreven, Beek, & Smeets (2015). However, according to Barlett (2007), cutoff
    frequencies between 4 and 8 Hz are often used in filtering movement data. Decided to use a wider range.
    :return: the optimal cutoff frequency
    """

    def autocorr(_signal):
        _signal_mean = np.mean(_signal)

        # Compute the autocorrelation using NumPy's correlate function
        _corr = np.correlate(_signal - _signal_mean, _signal - _signal_mean, mode='full')

        # Normalize the autocorrelation
        _corr /= (np.std(_signal) ** 2) * len(_signal)

        # based on Cappello, La Palombara, & Leardini (1996), only use up to 10 lags
        _corr = _corr[:10]

        # get the summed squares of the autocorrelation
        _corr = np.sum(_corr ** 2)

        return _corr

    norm_resid = []

    # compute the normalized residual autocorrelation for each cutoff frequency
    for fc in fc_test:
        # compute the residual
        resid = signal - low_butter(signal, fs, fc)
        # compute the normalized residual autocorrelation
        resid_autocorr = autocorr(resid)

        norm_resid.append(resid_autocorr)
    norm_resid = np.array(norm_resid)

    # return the cutoff frequency with the smallest normalized residual autocorrelation
    return fc_test[np.argmin(norm_resid)]


def rotation_mat_2d(angle):
    return np.array([
        [np.cos(angle), -np.sin(angle)],
        [np.sin(angle), np.cos(angle)]])

def compute_transformation_2d(start_pos: np.ndarray,
                              end_pos: np.ndarray,
                              to_dir: np.ndarray = np.array([0, 1]),):
    # normalized vector between the start and end positions
    movement_ax = end_pos - start_pos
    movement_ax_norm = movement_ax / np.linalg.norm(movement_ax)

    # get the angle between the desired direction and the movement axis
    angle = np.arccos(np.matmul(movement_ax_norm, to_dir))

    # get the rotational matrix based on the angle
    return rotation_mat_2d(angle)

def compute_transformation_3d(surface_points_x: np.ndarray,
                              surface_points_y: np.ndarray,
                              surface_points_z: np.ndarray,
                              horizontal_norm: Union[str, int] = 'y',
                              primary_ax: Union[str, int] = 'z',
                              secondary_ax: Union[str, int] = 'x',
                              full_output: bool = False) -> (Rotation, np.ndarray, dict):
    """
    :param surface_points_x: a vector specifying the x coordinates of points sampled from a surface
    :param surface_points_y: a vector specifying the y coordinates of points sampled from a surface
    :param surface_points_z: a vector specifying the z coordinates of points sampled from a surface
    :param horizontal_norm: the name of the axis that is perpendicular to the horizontal plane
    :param primary_ax: the name of the axis that corresponds to the primary movement direction
    :param secondary_ax: the name of the axis that corresponds to the secondary movement direction
    :param full_output: whether to return full output, which includes the objects for the plane and corners
    :return:
    """
    def find_unit_ax(ax: Union[str, int]):
        if (ax == 'x') | (ax == 0):
            return Vector(vg.basis.x), 0
        elif (ax == 'y') | (ax == 1):
            return Vector(vg.basis.y), 1
        elif (ax == 'z') | (ax == 2):
            return Vector(vg.basis.z), 2
        else:
            raise ValueError('the axis has to be "x", "y", or "z"!')

    def rotate_vec_to_target(_vec, _target):
        # Normalize the input normal vector
        _vec = _vec / np.linalg.norm(_vec)

        # Calculate the axis of rotation (cross product between normal and target)
        axis = np.cross(_vec, _target)

        # Calculate the cosine of the angle between the two vectors
        cos_angle = np.dot(_vec, _target)

        # Calculate the skew-symmetric cross-product matrix
        cross_product_matrix = np.array([
            [0, -axis[2], axis[1]],
            [axis[2], 0, -axis[0]],
            [-axis[1], axis[0], 0]
        ])

        # Create the rotation matrix using the Rodrigues' formula
        rotation_matrix = np.identity(3) + cross_product_matrix + np.dot(cross_product_matrix, cross_product_matrix) * (
                    1 - cos_angle) / (np.linalg.norm(axis) ** 2)

        return rotation_matrix

    def project_vector_onto_plane(_vec, _plane_normal):
        # Normalize the plane normal vector
        _plane_normal = _plane_normal / np.linalg.norm(_plane_normal)

        # Calculate the projection of the vector onto the plane
        projection = _vec - np.dot(_vec, _plane_normal) * _plane_normal

        return projection

    primary_norm, _ = find_unit_ax(primary_ax)
    secondary_norm, _ = find_unit_ax(secondary_ax)
    horizontal_norm, horizontal_ind = find_unit_ax(horizontal_norm)

    # concatenate the points
    surface_points = np.concatenate([np.expand_dims(surface_points_x, axis=1),
                                     np.expand_dims(surface_points_y, axis=1),
                                     np.expand_dims(surface_points_z, axis=1)], axis=1)

    # center the points before fitting a plane through it
    surface_center = np.mean(surface_points, axis=0)
    surface_points = surface_points - surface_center

    # find the best fitting plane and its surface normal
    surface_plane = Plane.best_fit(surface_points)
    surface_norm = Vector(surface_plane.cartesian()[:3])
    surface_norm = surface_norm / np.linalg.norm(surface_norm)

    # project the current surface normal to the horizontal plane
    screen_norm_ground = project_vector_onto_plane(surface_norm, horizontal_norm)

    # rotation to transform the surface normal to the horizontal plane
    rotation_horizontal = Rotation.from_matrix(rotate_vec_to_target(surface_norm, horizontal_norm))

    # rotation to align the surface orientation to the primary axis
    rotation_primary = Rotation.from_matrix(rotate_vec_to_target(screen_norm_ground, primary_norm))

    # combine the two rotations
    rotation = rotation_primary * rotation_horizontal

    # rotate the surface points to transform them back to horizontal
    surface_points_rot = Points(rotation.apply(surface_points)) + surface_center
    screen_plane_rot = Plane.best_fit(surface_points_rot)

    if full_output:
        return rotation, surface_center, {
            'surface_plane': surface_plane,
            'surface_points_rot': surface_points_rot,
            'surface_plane_rot': screen_plane_rot,
        }
    else:
        return rotation, surface_center


def rotate_coord(x: np.ndarray, y: np.ndarray, z: np.ndarray,
                 surface_center: np.ndarray, rotation: Rotation) -> (np.ndarray, np.ndarray, np.ndarray):
    """
    :param x: x coordinate
    :param y: y coordinate
    :param z: z coordinate
    :param surface_center: the center of the surface, obtained through compute_transformation
    :param rotation: the Rotation object obtained through compute_transformation
    :return: rotated coordinates
    """

    coord = np.concatenate([np.expand_dims(x, axis=1),
                            np.expand_dims(y, axis=1),
                            np.expand_dims(z, axis=1)], axis=1)

    coord -= surface_center  # center the points before applying the rotation
    coord_rot = rotation.apply(coord)
    return coord_rot[:, 0], coord_rot[:, 1], coord_rot[:, 2]


def find_movement_bounds(movement_feature: np.ndarray,
                         feature_threshold: Union[float, int],
                         allow_multiple_segments=False) -> (int, int):
    """
    :param movement_feature: a numpy ndarray (1D or 2D) that will be used to evaluate against a threshold. This could
    be velocity or acceleration. If the input velocity is 2D, then the Pythagorean of the velocities across different
    dimensions will be computed and used to determine movement.
    :param feature_threshold: the threshold based on which the movement initiation and termination will be determined.
    :param allow_multiple_segments: whether returning more than one segments or automatically identify the segment with
    the most data points
    :return: the movement initiation and termination indices, either a single index for each or a list of indices

    Find movement start and end indices.
    """
    dim = movement_feature.shape

    if len(dim) == 1:
        # only one coordinate was supplied, thus taking the absolute value of the velocity. This is to prevent
        # situations when negative velocity is not considered as exceeding the velocity threshold.
        feature_eval = np.abs(movement_feature)
    elif len(dim) == 2:
        # when more than one coordinates were supplied, use the pythagorean of all the supplied coordinates
        feature_eval = np.linalg.norm(movement_feature, axis=int(np.argmin(dim)))
    else:
        raise ValueError("The input velocity has to be a 1D or a 2D Numpy array!")

    n_frames = len(feature_eval)

    feature_threshold_ind = np.where(feature_eval >= feature_threshold)[0]

    if len(feature_threshold_ind) == 0:
        # in case there's no movement detected
        return np.array([np.nan, np.nan])
    else:
        # see if more than one movement segment is identified
        feature_ind = consecutive(feature_threshold_ind)

        # directly return the indices if there is only one segments
        if len(feature_ind) == 1:
            move_start_ind = feature_threshold_ind[0] - 1 if feature_threshold_ind[0] - 1 > 0 else 0
            move_end_ind = feature_threshold_ind[-1] + 1 if feature_threshold_ind[-1] + 1 < n_frames - 1 else n_frames - 1
            return move_start_ind, move_end_ind

        if allow_multiple_segments:
            # when there are more than one segments and the user wants to save all segments
            move_start_ind = []
            move_end_ind = []

            for segment in feature_ind:
                temp_start = segment[0]
                temp_end = segment[-1]

                move_start_ind.append(temp_start - 1 if temp_start > 0 else 0)
                move_end_ind.append(temp_end + 1 if temp_end < n_frames - 1 else n_frames - 1)
        else:
            # when there are more than one segments and the user wants to keep the segments with the most data
            feature_len = [len(feature) for feature in feature_ind]
            # only use the portion of movement with the largest number of samples
            max_feature = np.where(feature_len == np.max(feature_len))[0][0]
            feature_threshold_ind = feature_ind[max_feature]

            move_start_ind = feature_threshold_ind[0] - 1 if feature_threshold_ind[0] - 1 > 0 else 0
            move_end_ind = feature_threshold_ind[-1] + 1 if feature_threshold_ind[-1] + 1 < n_frames - 1 else n_frames - 1

        return move_start_ind, move_end_ind


def find_movement_bounds_percent_threshold(feature, percent_feature=0.05, allow_multiple_segments=False):
    """ Percentage of peak velocity as the threshold to find movement bounds. """
    peak_vel = np.max(np.abs(feature))
    vel_threshold = percent_feature * peak_vel
    return find_movement_bounds(feature, vel_threshold, allow_multiple_segments)


def find_movement_bounds_displacement(displacement: np.ndarray,
                                      pos_start: np.ndarray,
                                      pos_end: np.ndarray,
                                      threshold: float = 0.1):
    # get the Euclidean distance between the displacement trajectory and start position
    dist_start = np.linalg.norm(displacement - pos_start, axis=1)
    # get the Euclidean distance between the displacement trajectory and end position
    dist_end = np.linalg.norm(displacement - pos_end, axis=1)

    # find the indices where the distance is smaller than the threshold
    start_ind = np.where(dist_start < threshold)[0]
    end_ind = np.where(dist_end < threshold)[0]

    # return the start and end indices
    return start_ind[0], end_ind[-1]

def find_start_end_pos(x: np.ndarray, y: np.ndarray, z: np.ndarray,
                       ind_start: int, ind_end: int, ind_buffer: int = 20) -> (np.ndarray, np.ndarray):
    """
    :param x: position along the x axis
    :param y: position along the y axis
    :param z: position along the z axis
    :param ind_start: the movement start index, obtained using find_movement_boudns function
    :param ind_end: the movement end index, obtained using find_movement_boudns function
    :param ind_buffer: number of coordinates to take before and after the start and end positions, respectively
    :return: start and end positions
    """

    x, y, z = check_input_coord(x, y, z)

    ind_start_low_bound = ind_start - ind_buffer if ind_start - ind_buffer >= 0 else 0
    ind_end_up_bound = ind_end + ind_buffer if ind_end + ind_buffer <= len(x) - 1 else len(x) - 1

    start_x = x[ind_start_low_bound:ind_start]
    start_y = y[ind_start_low_bound:ind_start]
    start_z = z[ind_start_low_bound:ind_start]
    mean_start = np.mean(np.array([start_x, start_y, start_z]), axis=1)

    end_x = x[ind_end:ind_end_up_bound]
    end_y = y[ind_end:ind_end_up_bound]
    end_z = z[ind_end:ind_end_up_bound]
    mean_end = np.mean(np.array([end_x, end_y, end_z]), axis=1)

    return mean_start, mean_end


def b_spline_fit_1d(time_vec, coord, n_fit, smooth=0., return_spline=False):
    """
    parameterize a single coordinate dimension

    :param time_vec: vector with the time stamps
    :param coord: coordinates from a single dimension
    :param n_fit: number of data points to be fitted
    :param smooth: smoothing factor
    :param return_spline: whether to also return the spline object
    :return: time_fit, trajectory_fit
    """
    tck = interpolate.splrep(time_vec, coord,
                             s=smooth,  # smoothing factor
                             k=3,  # degree of the spline fit. Scipy recommends using cubic splines.
                             )
    time_fit = np.linspace(np.min(time_vec), np.max(time_vec), n_fit)
    spline = interpolate.BSpline(tck[0], tck[1], tck[2])

    if return_spline:
        return time_fit, spline(time_fit), spline
    else:
        return time_fit, spline(time_fit)


class Preprocesses(Enum):
    FILL_MISSING = 1
    SPATIAL_TRANSFORM = 2
    LOW_BUTTER = 3
    CENT_DIFF = 4


def composite_function(*functions):
    def compose(f, g):
        return lambda x: f(g(x))

    return reduce(compose, functions, lambda x: x)



"""
A class for 2D trajectory analysis.

This class inherits from the base class TrajectoryBase. It implements the abstract methods in the base class, and
provides concrete implementations for the abstract methods in the base class.

Written by X.M. Wang.

Wang, X.M., & Welsh, T.N. (2023). TAT-HUM: Trajectory Analysis Toolkit for Human Movements in Python.
"""

from tathum.trajectory_base import TrajectoryBase
from tathum.functions import *
import typing
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


class Trajectory2D(TrajectoryBase):

    N_DIM = 2  # 2D trajectory, needed for the parent class method validate_size()

    def __init__(self,
                 x: np.ndarray,
                 y: np.ndarray,

                 displacement_preprocess: tuple[Preprocesses, ...] = (Preprocesses.LOW_BUTTER, ),
                 velocity_preprocess: tuple[Preprocesses, ...] = (Preprocesses.CENT_DIFF, ),
                 acceleration_preprocess: tuple[Preprocesses, ...] = (Preprocesses.CENT_DIFF, ),

                 transform_end_points: tuple[np.ndarray, np.ndarray] = None,
                 transform_to: np.ndarray = None,

                 time: typing.Optional[np.ndarray] = None,

                 primary_dir: str = 'y',  # the primary direction of the movement

                 unit: str = 'mm',
                 missing_data_value: float = 0.,
                 fs: typing.Optional[int] = None, fc: typing.Optional[int] = None,
                 vel_threshold: float = 50.,

                 movement_selection_ax: str = 'xy',
                 movement_selection_method: str = 'length',
                 movement_selection_sign: str = 'positive',
                 custom_compute_movement_boundary: typing.Optional[callable] = None,

                 movement_pos_time_cutoff: float = 0.2,

                 spline_order: int = 3, n_spline_fit: int = 100,
                 ):

        self.x_original, self.y_original = x.copy(), y.copy()  # store a copy of the original data
        self.x, self.y = x, y
        self.x_vel, self.y_vel = np.array([]), np.array([])
        self.x_acc, self.y_acc = np.array([]), np.array([])
        self.n_frames = self.validate_size(n_dim=self.N_DIM)
        self.primary_dir = primary_dir

        if (time is None) & (fs is None):
            raise ValueError('You have to either specify the time stamps or the sampling frequency!')
        if time is None:
            self.time = np.linspace(0, self.n_frames * 1 / fs, num=self.n_frames, endpoint=False)
            self.time_original = self.time.copy()
        else:
            self.time_original = time.copy()
            self.time = time
            fs = 1 / np.mean(np.diff(self.time))
            if len(self.time) != self.n_frames:
                raise ValueError('The size of the input time stamps is not the same as the size of the coordinates!')

        # initialize the base class
        super().__init__(
            unit=unit,
            missing_data_value=missing_data_value,
            fs=fs, fc=fc,
            vel_threshold=vel_threshold,
            movement_selection_method=movement_selection_method, movement_selection_sign=movement_selection_sign,
            custom_compute_movement_boundary=custom_compute_movement_boundary,
            spline_order=spline_order, n_spline_fit=n_spline_fit, )

        # fill in missing data before performing spatial transformation (otherwise the missing data value would be
        # transformed as well)
        self.contain_missing, self.n_missing, self.ind_missing = self.missing_data()

        # if the cutoff frequency is not specified, then it will be computed automatically
        if self.fc is None:
            self.fc = find_optimal_cutoff_frequency(self.x, self.fs)

        # transform data if needed
        self.transform_end_points = transform_end_points
        if self.transform_end_points is not None:
            self.transform_to = transform_to if transform_to is not None else np.array([0, 1])
            start_pos, end_pos = self.transform_end_points[0], self.transform_end_points[1]
            self.transform_mat, self.transform_origin = self.compute_transform(start_pos, end_pos, self.transform_to)
            self.x, self.y = self.transform_data(self.x, self.y)
            self.x_original, self.y_original = self.transform_data(self.x_original, self.y_original)

        self.preprocess('displacement', displacement_preprocess)
        self.preprocess('velocity', velocity_preprocess)
        self.preprocess('acceleration', acceleration_preprocess)

        self.movement_displacement = self.find_movement_displacement(movement_selection_ax=movement_selection_ax)
        self.movement_velocity = self.find_movement_velocity(movement_selection_ax=movement_selection_ax)
        self.movement_acceleration = self.find_movement_acceleration(movement_selection_ax=movement_selection_ax)

        self.start_time, self.end_time, self.movement_ind = self.compute_movement_boundaries()
        self.contain_movement = self.validate_movement()

        if self.contain_movement:
            self.rt = self.start_time
            self.mt = self.end_time - self.start_time
            self.start_pos, self.end_pos = self.find_start_and_end_pos(time_cutoff=movement_pos_time_cutoff)

            self.time_movement = self.time[self.movement_ind]
            self.x_movement = self.x[self.movement_ind]
            self.y_movement = self.y[self.movement_ind]
            self.x_vel_movement = self.x_vel[self.movement_ind]
            self.y_vel_movement = self.y_vel[self.movement_ind]
            self.x_acc_movement = self.x_acc[self.movement_ind]
            self.y_acc_movement = self.y_acc[self.movement_ind]

            # trim the movement velocity to find the peak velocity
            self.movement_velocity = self.movement_velocity[self.movement_ind]
            peak_vel_idx = np.argmax(self.movement_velocity)
            self.peak_vel = self.movement_velocity[peak_vel_idx]
            self.time_to_peak_vel = self.time_movement[peak_vel_idx] - self.time_movement[0]
            self.time_after_peak_vel = self.time_movement[-1] - self.time_movement[peak_vel_idx]

            self.movement_acceleration = self.movement_acceleration[self.movement_ind]
            peak_acc_idx = np.argmax(self.movement_acceleration)
            self.peak_acc = self.movement_acceleration[peak_acc_idx]
            self.time_to_peak_acc = self.time_movement[peak_acc_idx] - self.time_movement[0]
            self.time_after_peak_acc = self.time_movement[-1] - self.time_movement[peak_acc_idx]

            self.time_fit, self.x_fit, self.x_spline = self.b_spline_fit_1d(self.time_movement, self.x_movement, self.n_spline_fit)
            _, self.y_fit, self.y_spline = self.b_spline_fit_1d(self.time_movement, self.y_movement, self.n_spline_fit)
            self.x_vel_fit = cent_diff(self.time_fit, self.x_fit)
            self.y_vel_fit = cent_diff(self.time_fit, self.y_fit)
            self.x_acc_fit = cent_diff(self.time_fit, self.x_vel_fit)
            self.y_acc_fit = cent_diff(self.time_fit, self.y_vel_fit)
        else:
            self.rt = None
            self.mt = None
            self.start_pos, self.end_pos = None, None

            self.time_movement = None
            self.x_movement = None
            self.y_movement = None
            self.x_vel_movement = None
            self.y_vel_movement = None
            self.x_acc_movement = None
            self.y_acc_movement = None

            self.vel_movement = None
            self.peak_vel = None
            self.time_to_peak_vel = None
            self.time_after_peak_vel = None

            self.movement_acceleration = None
            self.peak_acc = None
            self.time_to_peak_acc = None
            self.time_after_peak_acc = None

            self.time_fit = None
            self.x_fit = None
            self.x_spline = None
            self.y_fit = None
            self.y_spline = None
            self.x_vel_fit = None
            self.y_vel_fit = None
            self.x_acc_fit = None
            self.y_acc_fit = None

    def assign_preprocess_function(self,
                                   preprocess_var: str,
                                   preprocess: Preprocesses,):
        """
        A concrete implementation of the base class TrajectoryBase.

        Crucially, this assigns the unique preprocess wrapper functions unique to the current class. See the wrapper
        functions' respective implementations for details.
        """
        preprocess_order = self.find_preprocess_order(preprocess_var)
        if preprocess == Preprocesses.LOW_BUTTER:
            return self.low_butter, (preprocess_order, )
        elif preprocess == Preprocesses.CENT_DIFF:
            return self.cent_diff, (preprocess_order, )  # return a tuple of arguments
        else:
            raise ValueError('The preprocess is not recognized!')

    def low_butter(self, low_butter_order: int = 1):
        """
        A wrapper function for the low_butter() from tathum.functions specific to the current Trajectory2D class.
        """
        if low_butter_order == 1:
            self.x = low_butter(self.x, self.fs, self.fc)
            self.y = low_butter(self.y, self.fs, self.fc)
        elif low_butter_order == 2:
            self.x_vel = low_butter(self.x_vel, self.fs, self.fc)
            self.y_vel = low_butter(self.y_vel, self.fs, self.fc)
        elif low_butter_order == 3:
            self.x_acc = low_butter(self.x_acc, self.fs, self.fc)
            self.y_acc = low_butter(self.y_acc, self.fs, self.fc)
        else:
            raise ValueError('The order of the low-pass Butterworth filter has to be either 1 (for displacement), 2 '
                             '(for velocity), or 3 (for acceleration)!')

    def cent_diff(self, cent_diff_order: int = 2):
        """
        A wrapper function for the cent_diff() from tathum.functions specific to the current Trajectory2D class.
        """
        if cent_diff_order == 2:
            self.x_vel = cent_diff(self.time, self.x)
            self.y_vel = cent_diff(self.time, self.y)

        elif cent_diff_order == 3:
            self.x_acc = cent_diff(self.time, self.x_vel)
            self.y_acc = cent_diff(self.time, self.y_vel)

        else:
            raise ValueError('The order of the central difference has to be either 1 (for velocity ) or 2 '
                             '(for acceleration!')

    def find_start_and_end_pos(self, time_cutoff: float = 0.2):
        """
        Implementation of the abstract method in the base class TrajectoryBase, specific to the current Trajectory2D
        class.
        """
        ind_start = (self.time > self.start_time - time_cutoff) & (self.time < self.start_time + time_cutoff)
        if np.any(ind_start):
            start_x = self.x[ind_start]
            start_y = self.y[ind_start]
            mean_start = np.nanmean(np.array([start_x, start_y]), axis=1)
        else:
            mean_start = np.empty((2,)) * np.nan

        ind_end = (self.time > self.end_time - time_cutoff) & (self.time < self.end_time + time_cutoff)
        if np.any(ind_end) > 0:
            end_x = self.x[ind_end]
            end_y = self.y[ind_end]
            mean_end = np.nanmean(np.array([end_x, end_y]), axis=1)
        else:
            mean_end = np.empty((2,)) * np.nan

        return mean_start, mean_end

    def transform_data(self, x, y):
        coord = np.concatenate([np.expand_dims(x, axis=0),
                                np.expand_dims(y, axis=0)], axis=0)
        coord_rot = np.matmul(self.transform_mat, coord - self.transform_origin)  # + self.transform_origin - leaving this out so that the trajectory is centered at the origin
        return coord_rot[0], coord_rot[1]

    def format_results(self):
        return pd.DataFrame({
            'contain_movement': self.contain_movement,
            'fs': self.fs,
            'fc': self.fc,
            'rt': self.rt,
            'mt': self.mt,
            'movement_dist': np.linalg.norm(self.end_pos - self.start_pos),
            'peak_vel': self.peak_vel,
            'time_to_peak_vel': self.time_to_peak_vel,
            'time_after_peak_vel': self.time_after_peak_vel,
            'peak_acc': self.peak_acc,
            'time_to_peak_acc': self.time_to_peak_acc,
            'time_after_peak_acc': self.time_after_peak_acc,
        }, index=[0])

    def missing_data(self):
        """
        A wrapper function for the fill_missing_data() from tathum.functions.
        """
        self.x, self.y, self.time, missing_info = fill_missing_data(
            x=self.x, y=self.y, time=self.time, missing_data_value=self.missing_data_value,
        )
        self.n_frames = self.validate_size(self.N_DIM)  # remember to update n_frames
        return missing_info['contain_missing'], missing_info['n_missing'], missing_info['missing_ind']

    @staticmethod
    def compute_transform(start_pos, end_pos, transform_to):
        """
        A wrapper for the compute_transformation_2d() from tathum.functions.
        """
        return compute_transformation_2d(start_pos, end_pos, transform_to), np.expand_dims(start_pos, 1)  # use the start position as the origin

    @property
    def transform_end_points(self):
        """
        The 2D end points used to determine the transformation matrix.
        """
        return self._transform_end_point

    @transform_end_points.setter
    def transform_end_points(self, value):
        if value is None:
            self._transform_end_point = None
        else:
            err_msg = 'The transformation end point has to be a tuple of two 2D Numpy arrays! \n'
            if type(value) is not tuple:
                raise ValueError(f'{err_msg}'
                                 'The supplied value is not a tuple! ')
            elif len(value) != 2:
                raise ValueError(f'{err_msg}'
                                 f'The supplied tuple has {len(value)} elements instead of 2!')
            elif np.any([not type(v) == np.ndarray for v in value]):
                raise ValueError(f'{err_msg}'
                                 f'The supplied tuple contains elements that are not Numpy arrays!')
            elif np.any([not v.shape == (2, ) for v in value]):
                raise ValueError(f'{err_msg}'
                                 f'The supplied tuple contains elements are of size {[v.shape for v in value]}!')

            self._transform_end_point = value

    def find_movement_angle(self, perc_of_movement=0.2):
        """
        Find the angle of the movement based on the first x% of the movement on movement_plane_ax and primary_dir.
        :param perc_of_movement: the percentage of the movement to consider when computing the movement angle
        :return: the movement angle
        """
        if self.contain_movement:
            end_idx = int(np.round(perc_of_movement * len(self.x_movement)))

            displacement_vector = np.concatenate([np.expand_dims(self.x, axis=1),
                                                  np.expand_dims(self.y, axis=1)], axis=1)

            start_pos = displacement_vector[0]
            end_pos = displacement_vector[end_idx]
            movement_vector = end_pos - start_pos

            primary_dir_vector = np.zeros((3,))
            primary_dir_vector[self.primary_dir] = 1

            movement_angle = np.arccos(np.dot(movement_vector, primary_dir_vector) / np.linalg.norm(movement_vector))
            return movement_angle
        else:
            return None

    def find_movement_displacement(self, movement_selection_ax: str = 'xy'):
        if movement_selection_ax == 'x':
            return self.x
        elif movement_selection_ax == 'y':
            return self.y
        elif movement_selection_ax == 'xy':
            return np.linalg.norm(np.concatenate([
                np.expand_dims(self.x, axis=1),
                np.expand_dims(self.y, axis=1),
            ], axis=1), axis=1)
        else:
            raise ValueError('The movement selection axis has to be either x, y, or xy!')

    def find_movement_velocity(self, movement_selection_ax: str = 'xy'):
        if movement_selection_ax == 'x':
            return self.x_vel
        elif movement_selection_ax == 'y':
            return self.y_vel
        elif movement_selection_ax == 'xy':
            return np.linalg.norm(np.concatenate([
                np.expand_dims(self.x_vel, axis=1),
                np.expand_dims(self.y_vel, axis=1),
            ], axis=1), axis=1)
        else:
            raise ValueError('The movement selection axis has to be either x, y, or xy!')

    def find_movement_acceleration(self, movement_selection_ax: str = 'xy'):
        if movement_selection_ax == 'x':
            return self.x_acc
        elif movement_selection_ax == 'y':
            return self.y_acc
        elif movement_selection_ax == 'xy':
            return np.linalg.norm(np.concatenate([
                np.expand_dims(self.x_acc, axis=1),
                np.expand_dims(self.y_acc, axis=1),
            ], axis=1), axis=1)
        else:
            raise ValueError('The movement selection axis has to be either x, y, or xy!')

    def debug_plots(self, fig=None, axs=None):

        if axs is None:
            fig, axs = plt.subplots(2, 1)
            # plt.tight_layout()

        axs[0].plot(self.time_original, self.x_original, label='x', linestyle=':')
        axs[0].plot(self.time_original, self.y_original, label='y', linestyle=':')
        axs[0].scatter(self.time_original[self.ind_missing], self.x_original[self.ind_missing], color='k')

        axs[0].plot(self.time, self.x, label='x', c='r')
        axs[0].plot(self.time, self.y, label='y', c='g')

        # if self.contain_missing:
        #     for i_seg, seg in enumerate(self.missing_segments_movement):
        #         n_missing = self.n_missing_segments_movement[i_seg]
        #         seg_mid = int(np.median(seg) - 1)
        #
        #         axs[0].text(self.time[seg_mid], np.min([self.x_original, self.y_original, self.z_original]),
        #                     f'{n_missing}', fontsize=12, )

        # axs[0].plot(self.time, np.ones((len(self.time), 1)) * self.start_pos[principal_dir],
        #             c='c', linestyle=':', linewidth=3)
        # axs[0].plot(self.time, np.ones((len(self.time), 1)) * self.end_pos[principal_dir],
        #                 c='m', linestyle=':', linewidth=3)

        axs[0].plot([self.start_time, self.start_time],
                    [np.min([self.x, self.y]),
                     np.max([self.x, self.y])], label='start', c='c')
        axs[0].plot([self.end_time, self.end_time],
                    [np.min([self.x, self.y]),
                     np.max([self.x, self.y])], label='end', c='m')
        axs[0].set_xlabel('Time (seconds)')
        axs[0].set_ylabel('Displacement')
        # axs[0].set_title(f'{self.n_missing} missing data')
        # axs[0].legend()

        axs[1].plot(self.time, self.x_vel, label='x', c='r')
        axs[1].plot(self.time, self.y_vel, label='y', c='g')
        axs[1].plot([self.start_time, self.start_time],
                    [np.min([self.x_vel, self.y_vel]),
                     np.max([self.x_vel, self.y_vel])], label='start', c='c')
        axs[1].plot([self.end_time, self.end_time],
                    [np.min([self.x_vel, self.y_vel]),
                     np.max([self.x_vel, self.y_vel])], label='end', c='m')

        axs[1].set_xlabel('Time (seconds)')
        axs[1].set_ylabel('Velocity')

        plt.subplots_adjust(top=0.90, hspace=0.38, left=0.12, bottom=0.12)
        return fig, axs




"""
The TrajectoryMean class which allows storing a collection of individual trajectories to derive the characteristics of
the mean trajectory.

Written by X.M. Wang.

Wang, X.M., & Welsh, T.N. (2023). TAT-HUM: Trajectory Analysis Toolkit for Human Movements in Python.
"""
from typing import Optional
import matplotlib.pyplot as plt
from .trajectory_base import TrajectoryBase
from .coord import Coord
import numpy as np


class TrajectoryMean:
    x = Coord()
    y = Coord()
    z = Coord()
    time = Coord()

    def __init__(self, exp_condition: Optional[dict] = None):
        self.all_trajectories = []
        self.exp_condition = exp_condition
        self.x_mean = None
        self.y_mean = None
        self.z_mean = None
        self.x_sd = None
        self.y_sd = None
        self.z_sd = None

    @property
    def n_trajectories(self):
        return len(self.all_trajectories)

    def add_trajectory(self, traj: TrajectoryBase):
        """
        Add a trajectory to the TrajectoryMean object.
        :param traj: The trajectory to add.
        :return: None
        """
        self.all_trajectories.append(traj)

    def compute_mean_trajectory(self,
                                traj_names=('x_fit', 'y_fit', 'z_fit'),
                                post_script=''):
        """
        Compute the mean trajectory for the current condition.
        :param traj_names: The names of the trajectory attributes to compute the mean trajectory.
        :param post_script: The post script to append to the mean trajectory attribute names, if needed.
        :return: None
        """
        for name in traj_names:
            coord_name = name[0]
            mean_name = f'{coord_name}{post_script}_mean'
            sd_name = f'{coord_name}{post_script}_sd'

            temp = None
            for ind, traj in enumerate(self.all_trajectories):
                temp_coord = traj.__getattribute__(name)

                # center the coordinate's starting position
                temp_coord -= temp_coord[0]

                if len(temp_coord.shape) < 2:
                    temp_coord = np.expand_dims(temp_coord, 1)

                if temp is None:
                    temp = temp_coord.copy()
                else:
                    temp = np.append(temp, temp_coord, axis=1)

            mean = np.mean(temp, axis=1)
            sd = np.std(temp, axis=1)  # / np.sqrt(temp.shape[1])
            self.__setattr__(mean_name, mean)
            self.__setattr__(sd_name, sd)

    def remove_trajectory(self, ind_list):
        """
        Remove trajectories from the TrajectoryMean object.
        :param ind_list: The indices of the trajectories to remove.
        :return: None
        """
        for ind in sorted(ind_list, reverse=True):  # loop in reverse order to not throw off the subsequent indices
            self.all_trajectories.pop(ind)

    def debug_plots_trajectory(self,
                               principal_dir='xy',
                               single_name_generic='_fit',
                               mean_name_generic='_mean',
                               fig=None, ax=None,
                               show_text=False,):
        """
        Plot the individual trajectories and the mean trajectory.
        :param principal_dir: The principal directions to plot the trajectories.
        :param single_name_generic: The generic name of the trajectory attributes to plot the individual trajectories.
        :param mean_name_generic: The generic name of the trajectory attributes to plot the mean trajectory.
        :param fig: The figure to plot the trajectories.
        :param ax: The axes to plot the trajectories.
        :param show_text: Whether to show the text on the plot.
        :return: The figure and axes.
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)

        for ind, trajectory in enumerate(self.all_trajectories):
            plt_x = trajectory.__getattribute__(f'{principal_dir[0]}{single_name_generic}')
            plt_y = trajectory.__getattribute__(f'{principal_dir[1]}{single_name_generic}')
            n_plot = len(plt_x)
            ax.plot(plt_x, plt_y, pickradius=5, alpha=.7)
            ax.text(plt_x[int(n_plot / 2)], plt_y[int(n_plot / 2)], str(ind))

        mean_x = self.__getattribute__(f'{principal_dir[0]}{mean_name_generic}')
        mean_y = self.__getattribute__(f'{principal_dir[1]}{mean_name_generic}')

        ax.plot(mean_x, mean_y, linewidth=8, alpha=.5, color='b', pickradius=2)

        ax.scatter(mean_x[0], mean_y[0], marker='o', color='g', s=100)
        ax.scatter(mean_x[-1], mean_y[-1], marker='o', color='r', s=100)

        if show_text:
            ax.text(mean_x[0], mean_y[0], 'START', color='g', fontsize=20)
            ax.text(mean_x[-1], mean_y[-1], 'END', color='r', fontsize=20)

        ax.set_xlabel('x displacement')
        ax.set_ylabel('y displacement')

        return fig, ax



"""
The Trajectory class that takes the raw trajectory data and automatically processes the data to derive relevant action
measures.

Written by X.M. Wang.

Wang, X.M., & Welsh, T.N. (2023). TAT-HUM: Trajectory Analysis Toolkit for Human Movements in Python.
"""

import typing
import pandas as pd
from .coord import Coord
import numpy as np
from skspatial.objects import Plane
import matplotlib.pyplot as plt
from .trajectory_base import TrajectoryBase
from .functions import Preprocesses, cent_diff, low_butter, fill_missing_data, find_optimal_cutoff_frequency, \
    compute_transformation_3d


class Trajectory(TrajectoryBase):
    """
    Automatically processes trajectory data.
    """
    x = Coord()
    y = Coord()
    z = Coord()
    time = Coord()

    def __init__(self,
                 x: np.ndarray,
                 y: np.ndarray,
                 z: np.ndarray,

                 displacement_preprocess: typing.Tuple[Preprocesses, ...] = (Preprocesses.LOW_BUTTER,),
                 velocity_preprocess: typing.Tuple[Preprocesses, ...] = (Preprocesses.CENT_DIFF,),
                 acceleration_preprocess: typing.Tuple[Preprocesses, ...] = (Preprocesses.CENT_DIFF,),

                 transform_end_points=None,  # end points used for spatial transformation

                 time: typing.Optional[np.ndarray] = None,

                 movement_plane_ax='xz',  # 2D plane that specifies the principal direction of the reach
                 primary_dir='z',  # the primary movement direction
                 ground_dir='y',  # the ground's normal direction

                 movement_selection_ax='z',
                 movement_selection_method='length',
                 movement_selection_sign=None,
                 custom_compute_movement_boundary: typing.Optional[callable] = None,

                 center_movement=True,  # whether to center the movement at the origin

                 unit: str = 'mm',
                 missing_data_value: float = 0.,
                 fs: typing.Optional[int] = None,
                 fc: typing.Optional[int] = None,
                 vel_threshold: float = 50.,

                 movement_pos_time_cutoff=0.2,  # used for finding the start and end positions

                 spline_order=3,
                 n_spline_fit=100,
                 ):
        """
        :param x: x coordinate
        :param y: y coordinate
        :param z: z coordinate

        :param displacement_preprocess: the preprocesses to apply to the displacement data. See functions.Preprocesses
        for the available preprocesses.
        :param velocity_preprocess: the preprocesses to apply to the velocity data. See functions.Preprocesses for the
        available preprocesses.
        :param acceleration_preprocess: the preprocesses to apply to the acceleration data. See functions.Preprocesses
        for the available preprocesses.

        :param transform_end_points: A minimum of three 3D points that specifies the plane on which the movement
        occurred. If provided, the best-fitting plane for these points will be computed and used to transform the
        trajectory so that the transformed movement would take place on a horizontal surface. This could be derived
        using the same Trajectory class.

        :param time: the corresponding time stamps

        :param movement_plane_ax: The 2D plane (xy, xz, or yz) of interest.
        :param primary_dir: The primary direction of the movement. The primary direction is the direction of the
        movement that is of most theoretical relevance to the experiment. For example, if the experiment is about
        reaching, then the primary direction is the direction of the reach. The secondary direction would be deduced
        based on the primary direction and the movement plane.
        :param ground_dir: The ground's normal direction (upright). This is used to determine the transformation matrix.

        :param movement_selection_ax: The axis ('x', 'y', 'z', 'xy', 'xz', 'yz', 'xyz') along which the movement
        selection should be performed.
        :param movement_selection_method: The method to select the movement segment in case extraneous movements are
        detected. Can be chosen from ('length', 'sign', ). 'length' simply select the movement segment with the most
        number of points, whereas 'sign' selects the movement segment based on movement direction. 'length' is more
        effective for small extraneous movements whereas 'sign' is more effective for movement reversals during the data
        collection period. See movement_selection_sign for more.
        :param movement_selection_sign: Can be chosen from ('positive', 'negative'), only applies to directional
        movement. The sign is determined by the difference between the end and start position.
        :param custom_compute_movement_boundary: A custom function to compute the movement initiation and termination
        boundaries. The function should take in the trajectory object and return a tuple of size 3 for
        1) time_start (float, np.float64), 2) time_end (float, np.float64),
        3) movement indices (list[(int, np.ind64)], np.ndarray[(int, np.ind64)])

        :param center_movement: whether to center the movement at the origin

        :param unit: position measure's unit
        :param missing_data_value: the filler value for missing data, default to be 0. This would be used for missing
        data interpolation.
        :param fs: sampling frequency (Hz). This can be left as None if the time stamps are provided.
        :param fc: cutoff frequency (for the Butterworth filter). This can be left as None if the cutoff frequency is
        to be computed automatically.
        :param vel_threshold: The velocity threshold to determine movement initiation and termination, in the same unit
        as the displacement data.

        :param movement_pos_time_cutoff: The amount of time before/after the movement initiation/termination to
        consider when computing the trajectory's end positions. Instead of using the position at a particular time,
        the end position is the average of all positions within this time cutoff range.

        :param spline_order: the order of the b-spline fit, default to 3
        :param n_spline_fit: number of fitted trajectory points when using bspline
        """
        self.x_original, self.y_original, self.z_original = x.copy(), y.copy(), z.copy()  # keep a separate copy of the original data
        self.x, self.y, self.z = x, y, z
        self.x_vel, self.y_vel, self.z_vel = np.array([]), np.array([]), np.array([])
        self.x_acc, self.y_acc, self.z_acc = np.array([]), np.array([]), np.array([])
        self.n_frames = self.validate_size()

        if (time is None) & (fs is None):
            raise ValueError('You have to either specify the time stamps or the sampling frequency!')
        if time is None:
            self.time = np.linspace(0, self.n_frames * 1 / fs, num=self.n_frames, endpoint=False)
            self.time_original = self.time.copy()
        else:
            self.time_original = time.copy()
            self.time = time
            fs = 1 / np.mean(np.diff(self.time))
            if len(self.time) != self.n_frames:
                raise ValueError('The size of the input time stamps is not the same as the size of the coordinates!')

        super().__init__(
            unit=unit,
            missing_data_value=missing_data_value,
            fs=fs, fc=fc,
            vel_threshold=vel_threshold,
            movement_selection_method=movement_selection_method, movement_selection_sign=movement_selection_sign,
            custom_compute_movement_boundary=custom_compute_movement_boundary,
            spline_order=spline_order, n_spline_fit=n_spline_fit, )

        self.center_movement = center_movement

        self.movement_plane_ax = movement_plane_ax
        self.secondary_dir = None  # placeholder - will be set automatically after setting primary dir
        self.primary_dir = primary_dir
        self.ground_plane = None  # placeholder - will be set automatically after setting ground_dir
        self.ground_dir = ground_dir

        self.movement_selection_method = movement_selection_method
        self.movement_selection_sign = movement_selection_sign
        if self.movement_selection_method == 'sign':
            if self.movement_selection_sign is None:
                raise ValueError('movement_selection_method is set to "sign" but movement_selection_sign was not specified!')
            if movement_selection_ax is None:
                self.movement_selection_dir = self.primary_dir
            else:
                if movement_selection_ax == 'x':
                    self.movement_selection_dir = 0
                elif movement_selection_ax == 'y':
                    self._primary_dir = 1
                elif movement_selection_ax == 'z':
                    self._primary_dir = 2
                else:
                    raise ValueError('Invalid movement_selection_dir! Please use the following: x, y, or z')

        self.contain_movement = True  # whether there was actual movement

        # eliminate missing data; need to do it before the transformation
        self.contain_missing, self.n_missing, self.ind_missing, self.missing_segments, self.n_missing_segments = self.missing_data()

        # if the cutoff frequency is not specified, then it will be computed automatically
        if self.fc is None:
            self.fc = find_optimal_cutoff_frequency(self.x, self.fs)

        # transform the data if necessary
        self.transform_end_points = transform_end_points
        if self.transform_end_points is not None:
            self.transform_mat, self.transform_origin, transform_info = self.compute_transformation(
                self.transform_end_points, full_output=True)
            self.screen_plane = transform_info['screen_plane']
            self.screen_corners_rot = transform_info['screen_corners_rot']
            self.screen_plane_rot = transform_info['screen_plane_rot']
            self.x, self.y, self.z = self.transform_data(self.x, self.y, self.z)
            self.x_original, self.y_original, self.z_original = self.transform_data(
                self.x_original, self.y_original, self.z_original)

        # preprocess the data and obtain temporal derivatives
        self.preprocess('displacement', displacement_preprocess)
        self.preprocess('velocity', velocity_preprocess)
        self.preprocess('acceleration', acceleration_preprocess)

        # identify movement initiation and termination
        self.movement_displacement = self.find_movement_displacement(movement_selection_ax=movement_selection_ax)
        self.movement_velocity = self.find_movement_velocity(movement_selection_ax=movement_selection_ax)
        self.movement_acceleration = self.find_movement_acceleration(movement_selection_ax=movement_selection_ax)
        self.start_time, self.end_time, self.movement_ind = self.compute_movement_boundaries()
        self.contain_movement = self.validate_movement()

        if self.contain_movement:
            # check if the missing data are in the movement segment
            self.missing_segments_movement = []
            if self.contain_missing:
                # check if any of the missing indices are in the movement segment
                self.ind_missing_movement = [i for i, value in enumerate(self.ind_missing)
                                             if self.movement_ind[0] <= value <= self.movement_ind[-1]]
                self.n_missing_movement = len(self.ind_missing_movement)

                # do the same for respective missing segments
                for seg in self.missing_segments:
                    if (self.movement_ind[0] <= seg[0] <= self.movement_ind[-1]) & \
                            (self.movement_ind[0] <= seg[-1] <= self.movement_ind[-1]):
                        self.missing_segments_movement.append(seg)
            else:
                self.ind_missing_movement = []
                self.n_missing_movement = 0

            self.n_missing_segments_movement = np.array([len(seg) for seg in self.missing_segments_movement])

            self.rt = self.start_time
            self.mt = self.end_time - self.start_time
            self.start_pos, self.end_pos = self.find_start_and_end_pos(time_cutoff=movement_pos_time_cutoff)

            self.time_movement = self.time[self.movement_ind]
            self.x_movement = self.x[self.movement_ind]
            self.y_movement = self.y[self.movement_ind]
            self.z_movement = self.z[self.movement_ind]
            self.x_vel_movement = self.x_vel[self.movement_ind]
            self.y_vel_movement = self.y_vel[self.movement_ind]
            self.z_vel_movement = self.z_vel[self.movement_ind]
            self.x_acc_movement = self.x_vel[self.movement_ind]
            self.y_acc_movement = self.y_vel[self.movement_ind]
            self.z_acc_movement = self.z_vel[self.movement_ind]

            self.movement_velocity = self.movement_velocity[self.movement_ind]
            peak_vel_idx = np.argmax(self.movement_velocity)
            self.peak_vel = self.movement_velocity[peak_vel_idx]
            self.time_to_peak_vel = self.time_movement[peak_vel_idx] - self.time_movement[0]
            self.time_after_peak_vel = self.time_movement[-1] - self.time_movement[peak_vel_idx]

            self.movement_acceleration = self.movement_acceleration[self.movement_ind]
            peak_acc_idx = np.argmax(self.movement_acceleration)
            self.peak_acc = self.movement_acceleration[peak_acc_idx]
            self.time_to_peak_acc = self.time_movement[peak_acc_idx] - self.time_movement[0]
            self.time_after_peak_acc = self.time_movement[-1] - self.time_movement[peak_acc_idx]

            if self.center_movement:
                self.start_pos -= self.start_pos
                self.end_pos -= self.start_pos
                self.x_movement -= self.start_pos[0]
                self.y_movement -= self.start_pos[1]
                self.z_movement -= self.start_pos[2]

            self.time_fit, self.x_fit, self.x_spline = self.b_spline_fit_1d(self.time_movement, self.x_movement, self.n_spline_fit)
            _, self.y_fit, self.y_spline = self.b_spline_fit_1d(self.time_movement, self.y_movement, self.n_spline_fit)
            _, self.z_fit, self.z_spline = self.b_spline_fit_1d(self.time_movement, self.z_movement, self.n_spline_fit)

            # instead of normalizing the velocity and accelerations, simply use cent_diff to directly compute them
            self.x_vel_fit = cent_diff(self.time_fit, self.x_fit)
            self.y_vel_fit = cent_diff(self.time_fit, self.y_fit)
            self.z_vel_fit = cent_diff(self.time_fit, self.z_fit)
            self.x_acc_fit = cent_diff(self.time_fit, self.x_vel_fit)
            self.y_acc_fit = cent_diff(self.time_fit, self.y_vel_fit)
            self.z_acc_fit = cent_diff(self.time_fit, self.z_vel_fit)
        else:
            # in case the trajectory does not satisfy the movement initiation/termination criteria, in cases such as
            # when the participant number moved during the data collection period
            self.contain_missing = None
            self.n_missing = None
            self.rt = None
            self.mt = None
            self.start_pos, self.end_pos = None, None

            self.x_movement = None
            self.y_movement = None
            self.z_movement = None
            self.x_vel_movement = None
            self.y_vel_movement = None
            self.z_vel_movement = None
            self.x_acc_movement = None
            self.y_acc_movement = None
            self.z_acc_movement = None

            self.peak_vel = None
            self.time_to_peak_vel = None
            self.time_after_peak_vel = None

            self.peak_acc = None
            self.time_to_peak_acc = None
            self.time_after_peak_acc = None

    def find_movement_angle(self, perc_of_movement=0.2):
        """
        Find the angle of the movement based on the first x% of the movement on movement_plane_ax and primary_dir.
        :param perc_of_movement: the percentage of the movement to consider when computing the movement angle
        :return: the movement angle
        """
        if self.contain_movement:
            end_idx = int(np.round(perc_of_movement * len(self.x_movement)))

            displacement_vector = self.get_displacement_vector(self.movement_plane_ax)

            start_pos = displacement_vector[0]
            end_pos = displacement_vector[end_idx]
            movement_vector = end_pos - start_pos

            primary_dir_vector = np.zeros((3,))
            primary_dir_vector[self.primary_dir] = 1

            movement_angle = np.arccos(np.dot(movement_vector, primary_dir_vector) / np.linalg.norm(movement_vector))
            return movement_angle
        else:
            return None

    def assign_preprocess_function(self,
                                   preprocess_var: str,
                                   preprocess: Preprocesses) -> (callable, tuple):
        preprocess_order = self.find_preprocess_order(preprocess_var)
        if preprocess == Preprocesses.LOW_BUTTER:
            return self.low_butter, (preprocess_order, )
        elif preprocess == Preprocesses.CENT_DIFF:
            return self.cent_diff, (preprocess_order, )
        else:
            raise ValueError('The preprocess is not recognized!')

    def low_butter(self, low_butter_order: int = 1):
        """
        A wrapper function for the low_butter() from tathum.functions specific to the current Trajectory2D class.
        """
        if low_butter_order == 1:
            self.x = low_butter(self.x, self.fs, self.fc)
            self.y = low_butter(self.y, self.fs, self.fc)
            self.z = low_butter(self.z, self.fs, self.fc)
        elif low_butter_order == 2:
            self.x_vel = low_butter(self.x_vel, self.fs, self.fc)
            self.y_vel = low_butter(self.y_vel, self.fs, self.fc)
            self.z_vel = low_butter(self.z_vel, self.fs, self.fc)
        elif low_butter_order == 3:
            self.x_acc = low_butter(self.x_acc, self.fs, self.fc)
            self.y_acc = low_butter(self.y_acc, self.fs, self.fc)
            self.z_acc = low_butter(self.z_acc, self.fs, self.fc)
        else:
            raise ValueError('The order of the low-pass Butterworth filter has to be either 1 (for displacement), 2 '
                             '(for velocity), or 3 (for acceleration)!')

    def cent_diff(self, cent_diff_order: int = 2):
        """
        A wrapper function for the cent_diff() from tathum.functions specific to the current Trajectory2D class.
        """
        if cent_diff_order == 2:
            self.x_vel = cent_diff(self.time, self.x)
            self.y_vel = cent_diff(self.time, self.y)
            self.z_vel = cent_diff(self.time, self.z)
        elif cent_diff_order == 3:
            self.x_acc = cent_diff(self.time, self.x_vel)
            self.y_acc = cent_diff(self.time, self.y_vel)
            self.z_acc = cent_diff(self.time, self.z_vel)
        else:
            raise ValueError('The order of the central difference has to be either 1 (for velocity ) or 2 '
                             '(for acceleration!')

    def get_displacement_vector(self, ax):
        """
        Get the displacement vector along the specified axis.
        :param ax: the axis ('x', 'y', 'z', 'xy', 'xz', 'yz', 'xyz') along which the displacement vector should be
        :return: the displacement vector
        """
        axis_mapping = {
            'x': self.x,
            'y': self.y,
            'z': self.z,
            'xy': np.concatenate([np.expand_dims(self.x, axis=1), np.expand_dims(self.y, axis=1)], axis=1),
            'xz': np.concatenate([np.expand_dims(self.x, axis=1), np.expand_dims(self.z, axis=1)], axis=1),
            'yz': np.concatenate([np.expand_dims(self.y, axis=1), np.expand_dims(self.z, axis=1)], axis=1),
            'xyz': np.concatenate([
                np.expand_dims(self.x, axis=1),
                np.expand_dims(self.y, axis=1),
                np.expand_dims(self.z, axis=1)], axis=1),
        }
        return axis_mapping[ax]

    def find_movement_displacement(self, movement_selection_ax: str = 'z'):
        """
        Find the displacement vectors based on which the movement selection will be performed.
        :param movement_selection_ax: the axis ('x', 'y', 'z', 'xy', 'xz', 'yz', 'xyz') along which the movement will
        be selected.
        :return: the displacement vectors
        """
        if movement_selection_ax in ['x', 'y', 'z']:
            return np.abs(self.get_displacement_vector(movement_selection_ax))
        elif movement_selection_ax in ['xy', 'xz', 'yz', 'xyz']:
            return np.linalg.norm(self.get_displacement_vector(movement_selection_ax), axis=1)
        else:
            raise ValueError('Invalid movement_selection_ax! Please use the following: x, y, z, xy, xz, yz, or xyz')

    def get_velocity_vector(self, ax):
        """
        Get the velocity vector along the specified axis.
        :param ax: the axis ('x', 'y', 'z', 'xy', 'xz', 'yz', 'xyz') along which the velocity vector should be
        :return: the velocity vector
        """
        axis_mapping = {
            'x': self.x_vel,
            'y': self.y_vel,
            'z': self.z_vel,
            'xy': np.concatenate([np.expand_dims(self.x_vel, axis=1), np.expand_dims(self.y_vel, axis=1)], axis=1),
            'xz': np.concatenate([np.expand_dims(self.x_vel, axis=1), np.expand_dims(self.z_vel, axis=1)], axis=1),
            'yz': np.concatenate([np.expand_dims(self.y_vel, axis=1), np.expand_dims(self.z_vel, axis=1)], axis=1),
            'xyz': np.concatenate([
                np.expand_dims(self.x_vel, axis=1),
                np.expand_dims(self.y_vel, axis=1),
                np.expand_dims(self.z_vel, axis=1)], axis=1),
        }
        return axis_mapping[ax]

    def find_movement_velocity(self, movement_selection_ax: str = 'z'):
        """
        Find the velocity vectors based on which the movement selection will be performed.
        :param movement_selection_ax: the axis ('x', 'y', 'z', 'xy', 'xz', 'yz', 'xyz') along which the movement will
        be selected.
        :return: the velocity vectors
        """
        if movement_selection_ax in ['x', 'y', 'z']:
            return np.abs(self.get_velocity_vector(movement_selection_ax))
        elif movement_selection_ax in ['xy', 'xz', 'yz', 'xyz']:
            return np.linalg.norm(self.get_velocity_vector(movement_selection_ax), axis=1)
        else:
            raise ValueError('Invalid movement_selection_ax! Please use the following: x, y, z, xy, xz, yz, or xyz')


    def get_acceleration_vector(self, ax):
        """
        Get the acceleration vector along the specified axis.
        :param ax: the axis ('x', 'y', 'z', 'xy', 'xz', 'yz', 'xyz') along which the acceleration vector should be
        :return: the acceleration vector
        """
        axis_mapping = {
            'x': self.x_acc,
            'y': self.y_acc,
            'z': self.z_acc,
            'xy': np.concatenate([np.expand_dims(self.x_acc, axis=1), np.expand_dims(self.y_acc, axis=1)], axis=1),
            'xz': np.concatenate([np.expand_dims(self.x_acc, axis=1), np.expand_dims(self.z_acc, axis=1)], axis=1),
            'yz': np.concatenate([np.expand_dims(self.y_acc, axis=1), np.expand_dims(self.z_acc, axis=1)], axis=1),
            'xyz': np.concatenate([
                np.expand_dims(self.x_acc, axis=1),
                np.expand_dims(self.y_acc, axis=1),
                np.expand_dims(self.z_acc, axis=1)], axis=1),
        }
        return axis_mapping[ax]

    def find_movement_acceleration(self, movement_selection_ax: str = 'z'):
        """
        Find the acceleration vectors based on which the movement selection will be performed.
        :param movement_selection_ax: the axis ('x', 'y', 'z', 'xy', 'xz', 'yz', 'xyz') along which the movement will
        be selected.
        :return: the acceleration vectors
        """
        if movement_selection_ax in ['x', 'y', 'z']:
            return np.abs(self.get_acceleration_vector(movement_selection_ax))
        elif movement_selection_ax in ['xy', 'xz', 'yz', 'xyz']:
            return np.linalg.norm(self.get_acceleration_vector(movement_selection_ax), axis=1)
        else:
            raise ValueError('Invalid movement_selection_ax! Please use the following: x, y, z, xy, xz, yz, or xyz')

    def compute_transformation(self, screen_corners, full_output=False):
        """
        :param screen_corners: the corners of the surface on which the movement was performed
        :param full_output: whether to return full output, which includes the objects for the plane and corners
        :return: the transformation matrix, the origin of the transformation, and the transformed screen corners
        """
        screen_center = np.mean(screen_corners, axis=0)
        rotation, surface_center, transform_info = compute_transformation_3d(
            screen_corners[:, 0], screen_corners[:, 1], screen_corners[:, 2],
            horizontal_norm=self.ground_dir,
            primary_ax=self.primary_dir,
            secondary_ax=self.secondary_dir,
            full_output=True,
        )

        if full_output:
            return rotation, screen_center, {
                'screen_plane': transform_info['surface_plane'],
                'screen_corners_rot': transform_info['surface_points_rot'],
                'screen_plane_rot': transform_info['surface_plane_rot'],
            }
        else:
            return rotation, screen_center

    def transform_data(self, x, y, z):
        """
        Spatially transform the trajectory so that the trajectory is on a flat, horizontal plane.
        :param x: x coordinate
        :param y: y coordinate
        :param z: z coordinate
        :return: transformed coordinates
        """
        coord = np.concatenate([np.expand_dims(x, axis=1),
                                np.expand_dims(y, axis=1),
                                np.expand_dims(z, axis=1)], axis=1)
        coord -= self.transform_origin
        coord_rot = self.transform_mat.apply(coord) + self.transform_origin
        return coord_rot[:, 0], coord_rot[:, 1], coord_rot[:, 2]

    def find_start_and_end_pos(self, time_cutoff):
        """
        Find the start and end positions of the movement.
        :param time_cutoff: The amount of time before/after the movement initiation/termination to consider when
        computing the trajectory's end positions. Instead of using the position at a particular time, the end position
        is the average of all positions within this time cutoff range.
        :return: mean_start, mean_end: the start and end positions
        """

        ind_start = (self.time < self.start_time + time_cutoff) & (self.time > self.start_time - time_cutoff)
        if np.any(ind_start):
            start_x = self.x[ind_start]
            start_y = self.y[ind_start]
            start_z = self.z[ind_start]
            mean_start = np.mean(np.array([start_x, start_y, start_z]), axis=1)
        else:
            mean_start = np.empty((3,)) * np.nan

        ind_end = (self.time < self.end_time + time_cutoff) & (self.time > self.end_time - time_cutoff)
        if np.any(ind_end) > 0:
            end_x = self.x[ind_end]
            end_y = self.y[ind_end]
            end_z = self.z[ind_end]
            mean_end = np.mean(np.array([end_x, end_y, end_z]), axis=1)
        else:
            mean_end = np.empty((3,)) * np.nan

        return mean_start, mean_end

    @staticmethod
    def consecutive(data, step_size=1):
        """ splits the missing data indices into chunks"""
        return np.split(data, np.where(np.diff(data) != step_size)[0] + 1)

    # def find_missing_data(self):
    #     return find_missing_data(self.x, self.missing_data_value)

    def missing_data(self):
        """
        Find if there is any outliers in the original movement trajectory. The outliers are detected by comparing each
        consecutive points' difference with a certain proportion of the overall range of motion.

        threshold: the threshold for the difference of x, y, and z, before and after the missing data block. If the
        difference is small, then we will use linear interpolation to fill in the gap. By default, the threshold for
        each axis is 1 mm/s.
        """
        self.x, self.y, self.z, self.time, missing_info = fill_missing_data(
            x=self.x, y=self.y, z=self.z, time=self.time, missing_data_value=self.missing_data_value)
        self.n_frames = self.validate_size()  # remember to update n_frames
        return missing_info['contain_missing'], missing_info['n_missing'], missing_info['missing_ind'], missing_info['missing_ind_segments'], missing_info['n_missing_segments']

    def debug_plots(self, fig=None, axs=None):
        """ Create a debug plot that shows displacement, velocity, acceleration, and XY trajectory"""
        if axs is None:
            fig, axs = plt.subplots(2, 1)
            # plt.tight_layout()

        axs[0].plot(self.time_original, self.x_original, label='x', linestyle=':')
        axs[0].plot(self.time_original, self.y_original, label='y', linestyle=':')
        axs[0].plot(self.time_original, self.z_original, label='z', linestyle=':')
        axs[0].scatter(self.time_original[self.ind_missing], self.x_original[self.ind_missing], color='k')

        axs[0].plot(self.time, self.x, label='x', c='r')
        axs[0].plot(self.time, self.y, label='y', c='g')
        axs[0].plot(self.time, self.z, label='z', c='b')

        if self.contain_missing:
            for i_seg, seg in enumerate(self.missing_segments_movement):
                n_missing = self.n_missing_segments_movement[i_seg]
                seg_mid = int(np.median(seg) - 1)

                axs[0].text(self.time[seg_mid], np.min([self.x_original, self.y_original, self.z_original]),
                            f'{n_missing}', fontsize=12,)

        if self.contain_movement:
            for principal_dir in self.movement_plane_ax:
                axs[0].plot(self.time, np.ones((len(self.time), 1)) * self.start_pos[principal_dir],
                            c='c', linestyle=':', linewidth=3)
                axs[0].plot(self.time, np.ones((len(self.time), 1)) * self.end_pos[principal_dir],
                            c='m', linestyle=':', linewidth=3)

            axs[0].plot([self.start_time, self.start_time],
                        [np.min([self.x, self.y, self.z]),
                         np.max([self.x, self.y, self.z])], label='start', c='c')
            axs[0].plot([self.end_time, self.end_time],
                        [np.min([self.x, self.y, self.z]),
                         np.max([self.x, self.y, self.z])], label='end', c='m')
        axs[0].set_xlabel('Time (seconds)')
        axs[0].set_ylabel('Displacement')
        # axs[0].set_title(f'{self.n_missing} missing data')
        # axs[0].legend()

        axs[1].plot(self.time, self.x_vel, label='x', c='r')
        axs[1].plot(self.time, self.y_vel, label='y', c='g')
        axs[1].plot(self.time, self.z_vel, label='z', c='b')
        axs[1].plot([self.start_time, self.start_time],
                    [np.min([self.x_vel, self.y_vel, self.z_vel]),
                     np.max([self.x_vel, self.y_vel, self.z_vel])], label='start', c='c')
        axs[1].plot([self.end_time, self.end_time],
                    [np.min([self.x_vel, self.y_vel, self.z_vel]),
                     np.max([self.x_vel, self.y_vel, self.z_vel])], label='end', c='m')

        axs[1].set_xlabel('Time (seconds)')
        axs[1].set_ylabel('Velocity')

        plt.subplots_adjust(top=0.90, hspace=0.38, left=0.12, bottom=0.12)
        return fig, axs

    def display_missing_info(self):
        if self.contain_missing:
            print(f'This trial contains {self.n_missing} missing data points.\n'
                  f'Among them, there are {len(self.missing_segments_movement)} segments occurred during the movement!\n'
                  f'The size of the missing segments are: {self.n_missing_segments_movement}\n')
        else:
            print(f'This trial does not contain any missing data!')

    def format_results(self):
        return pd.DataFrame({
            'contain_movement': self.contain_movement,
            'fs': self.fs,
            'fc': self.fc,
            'rt': self.rt,
            'mt': self.mt,
            'movement_dist': np.linalg.norm(self.end_pos - self.start_pos),
            'peak_vel': self.peak_vel,
            'time_to_peak_vel': self.time_to_peak_vel,
            'time_after_peak_vel': self.time_after_peak_vel,
            'peak_acc': self.peak_acc,
            'time_to_peak_acc': self.time_to_peak_acc,
            'time_after_peak_acc': self.time_after_peak_acc,
        }, index=[0])

    def display_results(self):
        _results = self.format_results()

        for col in _results.columns:
            print(f'{col}: {_results[col].values[0]:.2f}')

    def demo_plots(self, fig=None, axs=None):
        if axs is None:
            fig, axs = plt.subplots(1, 1)
            # plt.tight_layout()

        axs.plot(self.time, self.x, label='x', linewidth=3, c='c')
        axs.plot(self.time, self.y, label='y', linewidth=3, c='m')
        axs.plot(self.time, self.z, label='z', linewidth=3, c='y')

        # axs.plot(self.time_original, self.x_original, label='x', linewidth=3, linestyle=':', c='k')
        # axs.plot(self.time_original, self.y_original, label='y', linewidth=3, linestyle=':', c='k')
        # axs.plot(self.time_original, self.z_original, label='z', linewidth=3, linestyle=':', c='k')
        # axs.scatter(self.time_original[self.ind_missing], self.x_original[self.ind_missing], color='k')

        axs.plot([self.start_time, self.start_time],
                 [np.min([self.x, self.y, self.z]),
                  np.max([self.x, self.y, self.z])],
                 label='Movement Start', color='g', linewidth=4, linestyle=':')
        axs.plot([self.end_time, self.end_time],
                 [np.min([self.x, self.y, self.z]),
                  np.max([self.x, self.y, self.z])],
                 label='Movement End', color='r', linewidth=4, linestyle=':')
        axs.set_xlabel('Time (s)', fontdict={'fontsize': 16})
        axs.set_ylabel('Displacement (mm)', fontdict={'fontsize': 16})
        axs.set_xticklabels(axs.get_xticklabels(), fontsize=12)
        axs.set_yticklabels(axs.get_yticklabels(), fontsize=12)
        # axs[0].set_title(f'Displacement, {self.n_missing} missing data')
        axs.legend()

        return fig, axs

    def debug_traj_3d(self, ax=None, color='k'):
        """ Create a debug plot that shows 3d trajectories"""
        if ax is None:
            plt.figure()
            ax = plt.axes(projection='3d')
        else:
            if ax.name != '3d':
                raise ValueError('The input MPL ax has to be a 3D axis! You can simply create an axis with the '
                                 'projection keyword: \n'
                                 'ax = plt.axes(projection=\'3d\')')

        # # ax.plot3D(self.x_smooth, self.y_smooth, self.z_smooth)
        # ax.scatter(self.x_smooth, self.y_smooth, self.z_smooth, c='k')
        # ax.scatter(self.x_smooth[0], self.y_smooth[0], self.z_smooth[0], c='r')  # beginning
        # ax.scatter(self.x_smooth[-1], self.y_smooth[-1], self.z_smooth[-1], c='g')  # end

        ax.plot3D(self.x, self.y, self.z, c=color)
        ax.scatter(self.x[0], self.y[0], self.z[0], c='r')  # beginning
        ax.scatter(self.x[-1], self.y[-1], self.z[-1], c='g')  # end

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')

        return ax

    def debug_traj_2d(self, x_axis, y_axis, var='', ax=None, center=True, full_output=False):
        """
        :param x_axis: the trajectory axis to be plotted on the x-axis of the 2d plot
        :param y_axis: the trajectory axis to be plotted on the y-axis of the 2d plot
        :param var: the variable to be plotted
        :param ax: matplotlib 2d axis object
        :param center: whether to center the data
        :param full_output: if True, will also return plot x and y coordinates
        :return: matplotlib 2d axis object
        """

        def check_axis(axis_name, _var):
            axis_name = axis_name.lower()

            if _var != '':
                _var = f'_{_var}'

            if (axis_name == 'x') | (axis_name == 'y') | (axis_name == 'z'):
                pos = getattr(self, f'{axis_name}{_var}')
            elif axis_name == 'xy':
                pos = np.sqrt((getattr(self, f'x{_var}') ** 2 +
                               getattr(self, f'y{_var}') ** 2))
            elif axis_name == 'xz':
                pos = np.sqrt((getattr(self, f'x{_var}') ** 2 +
                               getattr(self, f'z{_var}') ** 2))
            elif axis_name == 'yz':
                pos = np.sqrt((getattr(self, f'y{_var}') ** 2 +
                               getattr(self, f'z{_var}') ** 2))
            else:
                raise ValueError("Unrecognized trajectory axis name! Only support 1D or 2D axes!")
            return pos - pos[0] if center else pos

        plot_x = check_axis(x_axis, var)
        plot_y = check_axis(y_axis, var)
        line = ax.plot(plot_x, plot_y, alpha=.5)
        ax.scatter(plot_x[0], plot_y[0], c='r')
        ax.scatter(plot_x[-1], plot_y[-1], c='g')

        if full_output:
            return ax, line, plot_x, plot_y
        else:
            return ax, line

    @property
    def movement_plane_ax(self):
        return self._movement_plane

    @movement_plane_ax.setter
    def movement_plane_ax(self, value):
        if value == 'xy':
            self._movement_plane = [0, 1]
        elif value == 'xz':
            self._movement_plane = [0, 2]
        elif value == 'yz':
            self._movement_plane = [1, 2]
        else:
            raise ValueError('Invalid principal directions! Please use the following: '
                             'xy, xz, or yz')

    @property
    def primary_dir(self):
        return self._primary_dir

    @primary_dir.setter
    def primary_dir(self, value):
        if value == 'x':
            self._primary_dir = 0
        elif value == 'y':
            self._primary_dir = 1
        elif value == 'z':
            self._primary_dir = 2
        else:
            raise ValueError('Invalid primary directions! Please use the following: '
                             'x, y, or z')

        self.secondary_dir = list(set(self.movement_plane_ax) - {self.primary_dir})[0]

    @property
    def ground_dir(self):
        return self._ground_dir

    @ground_dir.setter
    def ground_dir(self, value):
        if value == 'x':
            self._ground_dir = 0
        elif value == 'y':
            self._ground_dir = 1
        elif value == 'z':
            self._ground_dir = 2
        else:
            raise ValueError('Invalid principal directions! Please use the following: '
                             'x, y, or z')

        ground_points = np.random.random((3, 3))
        ground_points[:, self._ground_dir] = 0
        self.ground_plane = Plane.best_fit(ground_points)



"""
This module contains the base class for all trajectory classes. It defines the common attributes and methods that are
shared by all trajectory classes. It also defines the abstract methods that all concrete classes need to implement.

Written by X.M. Wang.

Wang, X.M., & Welsh, T.N. (2023). TAT-HUM: Trajectory Analysis Toolkit for Human Movements in Python.
"""

from abc import ABC, abstractmethod
from tathum.coord import Coord
from tathum.functions import *


class TrajectoryBase(ABC):
    # pre-define the coordinates to be used for all concrete classes
    x = Coord()
    y = Coord()
    z = Coord()
    time = Coord()

    def __init__(self,
                 unit: str = 'mm',
                 n_dim: int = 3,
                 missing_data_value: float = 0.,
                 fs: int = 250, fc: int = 10,
                 vel_threshold: float = 50.,
                 movement_selection_method: str = 'length', movement_selection_sign: str = 'positive',
                 custom_compute_movement_boundary: callable = None,
                 spline_order: int = 3, n_spline_fit: int = 100,
                 ):
        self.vel_threshold = vel_threshold
        self.unit = unit
        self.n_dim = n_dim
        self.fs = fs
        self.fc = fc
        self.missing_data_value = missing_data_value
        self.spline_order = spline_order
        self.n_spline_fit = n_spline_fit
        self.movement_selection_method = movement_selection_method
        self.movement_selection_sign = movement_selection_sign
        self.custom_compute_movement_boundary = custom_compute_movement_boundary
        self.start_time, self.end_time, self.movement_ind = np.nan, np.nan, np.nan
        self.start_pos, self.end_pos = np.nan, np.nan
        self.movement_displacement, self.movement_velocity = np.nan, np.nan
        self.contain_movement = False

    @property
    def n_frames(self):
        """
        The number of frames in the trajectory.
        """
        return self._n_frames

    @n_frames.setter
    def n_frames(self, value):
        self._n_frames = value

    @property
    def movement_velocity(self):
        """
        The velocity coordinate used to determine the movement boundaries.
        """
        return self._movement_velocity

    @movement_velocity.setter
    def movement_velocity(self, value):
        self._movement_velocity = value

    @property
    def movement_displacement(self):
        """
        The displacement coordinate used to determine the movement boundaries.
        """
        return self._movement_displacement

    @movement_displacement.setter
    def movement_displacement(self, value):
        self._movement_displacement = value

    def preprocess(self, preprocess_var: str, preprocess_tuple: tuple):
        """
        A generic method that goes through a tuple of preprocesses and assign the corresponding function and relevant
        input.
        :param preprocess_var: The variable to be preprocessed
        :param preprocess_tuple: The tuple of preprocesses
        """
        for p in preprocess_tuple:
            f, val = self.assign_preprocess_function(preprocess_var, p)
            f() if val is None else f(*val)

    @abstractmethod
    def assign_preprocess_function(self, preprocess_var: str, preprocess: Preprocesses) -> (callable, tuple):
        """
        An abstract method that all concrete classes need to implement. This method assigns the preprocess function
        unique to the concrete class and the relevant variable to be preprocessed.
        :param preprocess_var: The variable to be preprocessed
        :param preprocess: The preprocess to be applied
        :return: The preprocess function and the relevant input
        """
        pass

    @abstractmethod
    def find_start_and_end_pos(self, time_cutoff: float):
        """
        Abstract method to be called after compute_movement_boundaries().

        Find the start and end positions of the movement.
        """
        pass

    @staticmethod
    def b_spline_fit_1d(time_vec, coord, n_fit, smooth=0., return_spline=True):
        """
        A static method to fit a B-spline to a 1D coordinate.

        This is simply an interface to the function of the same name in tathum.functions.
        """
        return b_spline_fit_1d(time_vec, coord, n_fit, smooth=smooth, return_spline=return_spline)

    def validate_size(self, n_dim: int = 3):
        """ Validate input coordinate size. """
        n_x, n_y = len(self.x), len(self.y)

        if n_dim == 3:
            n_z = len(self.z)
            if not (n_x == n_y == n_z):
                raise ValueError("The input x, y, and z have to be of the same size! \n"
                                 f"Instead, len(x)={len(self.x)}, len(y)={len(self.y)}, len(z)={len(self.z)}")
        elif n_dim == 2:
            if not (n_x == n_y):
                raise ValueError("The input x and y have to be of the same size! \n"
                                 f"Instead, len(x)={len(self.x)}, len(y)={len(self.y)}")
        return n_x

    def compute_movement_boundaries(self):
        """
        Computes the movement boundaries based on the velocity profile.
        """
        output_err_msg = 'The custom movement boundary method must return a tuple of size 3 for ' \
                         'time_start (float, np.float64), ' \
                         'time_end (float, np.float64), ' \
                         'movement indices (list[(int, np.ind64)], np.ndarray[(int, np.ind64)])!! \n'
        if self.custom_compute_movement_boundary is not None:

            boundary_output = self.custom_compute_movement_boundary(self)

            if boundary_output is None:
                raise ValueError(f'{output_err_msg}'
                                 'Current return value is None!')

            if type(boundary_output) is not tuple:
                raise ValueError(f'{output_err_msg}'
                                 f'Current return data type is {type(boundary_output)}!')

            if len(boundary_output) != 3:
                raise ValueError(f'{output_err_msg}'
                                 f'Current return size is {len(boundary_output)}!')

            if not isinstance(boundary_output[0], (float, np.float64)):
                raise ValueError(f'{output_err_msg}'
                                 f'Current return data type for time_start is {type(boundary_output[0])}!')

            if not isinstance(boundary_output[1], (float, np.float64)):
                raise ValueError(f'{output_err_msg}'
                                 f'Current return data type for time_end is {type(boundary_output[1])}!')

            if not isinstance(boundary_output[2], (list, np.ndarray)):
                raise ValueError(f'{output_err_msg}'
                                 f'Current return data type for vel_threshold_ind is {type(boundary_output[2])}!')

            if not all(isinstance(item, (int, np.int64)) for item in boundary_output[2]):
                raise ValueError(f'{output_err_msg}'
                                 f'Current return data type for values in movement indices are not all int!')

            return boundary_output[0], boundary_output[1], boundary_output[2]
        else:
            vel_threshold_ind = np.where(self.movement_velocity >= self.vel_threshold)[0]

            if len(vel_threshold_ind) == 0:
                # in case there's no movement detected
                return np.nan, np.nan, np.nan

            vel_ind = consecutive(vel_threshold_ind)

            # in case there are multiple crossings at the threshold velocity
            if len(vel_ind) > 1:
                if self.movement_selection_method == 'length':
                    # only use the portion of movement with the largest number of samples
                    vel_len = [len(vel) for vel in vel_ind]
                    max_vel = np.where(vel_len == np.max(vel_len))[0][0]
                    vel_threshold_ind = vel_ind[max_vel]

                elif self.movement_selection_method == 'sign':
                    movement_dist = []
                    for vel in vel_ind:
                        movement_dist.append(self.movement_displacement[vel[-1]] - self.movement_displacement[vel[0]])
                    movement_dist = np.array(movement_dist)

                    if self.movement_selection_sign == 'positive':
                        segment_id = np.where(movement_dist > 0)[0]
                    else:
                        segment_id = np.where(movement_dist < 0)[0]

                    if len(segment_id) == 1:
                        vel_threshold_ind = vel_ind[segment_id[0]]
                    elif len(segment_id) > 1:
                        # get the segment with the most data
                        vel_len = [len(vel_ind[seg_id]) for seg_id in segment_id]
                        max_vel = np.where(vel_len == np.max(vel_len))[0][0]
                        vel_threshold_ind = vel_ind[segment_id[max_vel]]
                    else:
                        print('no valid movement detected!')
                        return np.nan, np.nan, np.nan

            move_start_ind = vel_threshold_ind[0] - 1 if vel_threshold_ind[0] > 0 else 0
            move_end_ind = vel_threshold_ind[-1] + 1 if vel_threshold_ind[-1] < self.n_frames - 1 else self.n_frames - 1

            time_start = self.time[move_start_ind]
            time_end = self.time[move_end_ind]

            return time_start, time_end, vel_threshold_ind

    def validate_movement(self):
        """
        Check to see if the trajectory contains valid movement
        """
        if np.isnan(self.start_time) | np.isnan(self.end_time):
            return False
        elif len(self.movement_ind) <= self.spline_order:  # when few data points are detected
            return False
        else:
            return True

    @staticmethod
    def find_preprocess_order(preprocess_var: str):
        if preprocess_var == 'displacement':
            return 1
        elif preprocess_var == 'velocity':
            return 2
        elif preprocess_var == 'acceleration':
            return 3
        else:
            raise ValueError('The preprocess variable has to be either displacement, velocity, or acceleration!')

    @abstractmethod
    def format_results(self):
        """
        Format the results into a Pandas DataFrame.
        """
        pass



"""
A custom class to store coordinate data.

Written by X.M. Wang.

Wang, X.M., & Welsh, T.N. (2023). TAT-HUM: Trajectory Analysis Toolkit for Human Movements in Python.

"""

import numpy as np
import pandas as pd


class Coord:
    """
    This is a descriptor class that stores the coordinates of a SINGLE trajectory dimension.
    """

    def __set_name__(self, owner, name):
        self.name = f'_{name}'

    def __get__(self, instance, owner):
        return getattr(instance, self.name)

    def __set__(self, instance, value):
        if value is None:
            setattr(instance, self.name, None)
        elif (type(value) is pd.Series) | (type(value) is pd.DataFrame) | (type(value) is np.ndarray):
            if type(value) is pd.Series:
                value = value.to_numpy(dtype=float)
            dim = value.shape

            if len(dim) == 1:
                setattr(instance, self.name, value)
            else:
                raise ValueError(f'The input coordinate array has to be 1D, instead it is {len(dim)}D!')
        else:
            raise ValueError('The input coordinates have to be either a Numpy array or Pandas Series! \n'
                             f'Instead, you used {type(value)}')




from setuptools import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='tathum',
    version='1.0.2',
    description='TAT-HUM: Trajectory Analysis Toolkit for Human Movement',
    url='https://github.com/xywang01/TAT-HUM',
    download_url='https://github.com/xywang01/TAT-HUM/archive/refs/tags/1.0.2.tar.gz',
    author='X. Michael Wang, Centre for Motor Control, University of Toronto, Faculty of Kinesiology and Physical Education',
    author_email='michaelwxy.wang@utoronto.ca',
    license='MIT',
    packages=['tathum', ],
    zip_safe=False,
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'typing',
        'numpy',
        'pandas',
        'scipy',
        'scikit-spatial',
        'vg',
        'pytransform3d',
        'matplotlib',
        'seaborn',
        'jupyter',
    ]
)
